// Copyright (c) 2012-2014 Rowley Associates Limited.
//
// This file may be distributed under the terms of the License Agreement
// provided with this software.
//
// THIS FILE IS PROVIDED AS IS WITH NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#include <ctl_api.h>
#ifndef TARGET_AM335x
#include "OMAP35x.h"
#define N 96
#else
#include "AM335x.h"
#define N 128
#endif

CTL_ISR_FN_t fnRAMVectors[N];

int
ctl_set_isr(unsigned int vector, 
            unsigned int priority, 
            CTL_ISR_TRIGGER_t trigger, 
            CTL_ISR_FN_t isr, 
            CTL_ISR_FN_t *oldisr)
{
  if (oldisr)
    *oldisr = fnRAMVectors[vector];
  fnRAMVectors[vector] = isr;
  *(volatile unsigned *)(INTC_BASE+0x100+vector*4) = (priority)<<2; // INTCPS_ILRm

  (void)trigger;
  return(0);
}

int
ctl_unmask_isr(unsigned int vector)
{
  *(volatile unsigned *)(INTC_BASE+0x88+(0x20*(vector/32))) = 1<<(vector % 32); // INTCPS_MIR_CLEARn
  return(0);
}

int 
ctl_mask_isr(unsigned int vector)
{
  *(volatile unsigned *)(INTC_BASE+0x8C+(0x20*(vector/32))) = 1<<(vector % 32); // INTCPS_MIR_SETn
  return(0);
}

int
ctl_lowest_isr_priority(void)
{
  return 0x3f;
}

int
ctl_highest_isr_priority(void)
{
  return 0;
}

int 
ctl_adjust_isr_priority(int priority, int n)
{
  return priority - n;
}

#ifdef TARGET_AM335x

#define SystemCoreClock 24000000
#define ONE_MS (SystemCoreClock/1000)
#define TEN_MS (ONE_MS*10)

static CTL_ISR_FN_t userTimerISR;

static void DMTimerIsr(void)
{
#ifdef CTL_TASKING
  ctl_time_increment = (0xffffffff-DMTIMER2_TLDR)/ONE_MS;
#endif
  userTimerISR();
  DMTIMER2_IRQSTATUS = (1<<1);
}

void ctl_start_timer(CTL_ISR_FN_t timerFn)
{    
  userTimerISR = timerFn;  
  
  //DMTimer2ModuleClkConfig();
  CM_DPLL_CLKSEL_TIMER2_CLK = 1; // M_OSC clock source
  CM_PER_TIMER2_CLKCTRL = 2; // MODULE ENABLE

  ctl_set_isr(68, ctl_adjust_isr_priority(ctl_lowest_isr_priority(), 1), 0, DMTimerIsr, 0);
  ctl_unmask_isr(68);

  DMTIMER2_TIOCP_CFG = 1;
  while ((DMTIMER2_TIOCP_CFG & 1)); 
  DMTIMER2_TLDR = 0xffffffff-TEN_MS;
  DMTIMER2_TTGR = 1;
  DMTIMER2_TCLR = 3;
  DMTIMER2_IRQENABLE_SET = 2;
}

#ifdef CTL_TASKING

void
ctl_sleep(unsigned delay)
{
  CTL_TIME_t sleep_time;
  unsigned sleep_priority;
  if (delay == 0 || delay > 10)
    {
      ctl_global_interrupts_disable();
      sleep_time = ctl_current_time;
      DMTIMER2_TCLR = 0;
      DMTIMER2_TLDR = 0xffffffff-(ONE_MS * delay);
      DMTIMER2_TTGR = 1;
      DMTIMER2_TCLR = 3;
      ctl_interrupt_count = 1;
      sleep_priority = ctl_task_set_priority(ctl_task_executing, 255);
      ctl_interrupt_count = 0;
      ctl_reschedule_on_last_isr_exit = 0;
      ctl_global_interrupts_enable();
    }
  __asm("wfi");
  if (delay == 0 || delay > 10)
    {
      ctl_global_interrupts_disable();
      DMTIMER2_TCLR = 0;
      if (ctl_current_time == sleep_time) // something else woke us up and didn't update the time
        {
          if (DMTIMER2_IRQSTATUS & (1<<1))
            ctl_time_increment = (0xffffffff-DMTIMER2_TCRR)/ONE_MS;
          else
            ctl_time_increment = (0xffffffff-DMTIMER2_TLDR)/ONE_MS;
          ctl_increment_tick_from_isr();
          ctl_reschedule_on_last_isr_exit = 0;
        }
      DMTIMER2_TLDR = 0xffffffff-TEN_MS;
      DMTIMER2_TTGR = 1;
      DMTIMER2_TCLR = 3;
      ctl_global_interrupts_enable();
      ctl_task_set_priority(ctl_task_executing, sleep_priority);
    }
}

void
ctl_woken()
{
}

#endif

#else

#define SystemCoreClock 13000000
#define ONE_MS (SystemCoreClock/1000)
#define TEN_MS (ONE_MS*10)

static CTL_ISR_FN_t userTimerISR;

static 
void GPT1_IRQHandler(void)
{
#ifdef CTL_TASKING
  ctl_time_increment = (0xffffffff-GPTIMER1_TLDR)/ONE_MS;
#endif
  userTimerISR();
  GPTIMER1_TISR = 2;
}

void 
ctl_start_timer(CTL_ISR_FN_t timerFn)
{    
  userTimerISR = timerFn;  

  // enable clocks to gptimer1
  CM_FCLKEN_PER |= (1<<3);
  CM_ICLKEN_PER |= (1<<3);

  // set gptimer1 irq
  ctl_set_isr(37, ctl_adjust_isr_priority(ctl_lowest_isr_priority(), 1), 0, GPT1_IRQHandler, 0);
  ctl_unmask_isr(37);

  // reset gptimer1
  GPTIMER1_TIOCP_CFG = (1<<1);
  while (!(GPTIMER1_TISTAT & 1)); 
  GPTIMER1_TLDR = 0xffffffff-TEN_MS;
  GPTIMER1_TTGR = 1;
  GPTIMER1_TCLR = 3;
  GPTIMER1_TIER = 2;
}

#ifdef CTL_TASKING

void
ctl_sleep(unsigned delay)
{
  CTL_TIME_t sleep_time;
  unsigned sleep_priority;
  if (delay == 0 || delay > 10)
    {
      ctl_global_interrupts_disable();
      sleep_time = ctl_current_time;
      GPTIMER1_TCLR = 0;
      GPTIMER1_TLDR = 0xffffffff-(ONE_MS * delay);
      GPTIMER1_TTGR = 1;
      GPTIMER1_TCLR = 3;
      ctl_interrupt_count = 1;
      sleep_priority = ctl_task_set_priority(ctl_task_executing, 255);
      ctl_interrupt_count = 0;
      ctl_reschedule_on_last_isr_exit = 0;
      ctl_global_interrupts_enable();
    }
  __asm("wfi");
  if (delay == 0 || delay > 10)
    {
      ctl_global_interrupts_disable();
      GPTIMER1_TCLR = 0;
      if (ctl_current_time == sleep_time) // something else woke us up and didn't update the time
        {
          if (GPTIMER1_TISR & (1<<1))
            ctl_time_increment = (0xffffffff-GPTIMER1_TCRR)/ONE_MS;
          else
            ctl_time_increment = (0xffffffff-GPTIMER1_TLDR)/ONE_MS;
          ctl_increment_tick_from_isr();
          ctl_reschedule_on_last_isr_exit = 0;
        }
      GPTIMER1_TLDR = 0xffffffff-TEN_MS;
      GPTIMER1_TTGR = 1;
      GPTIMER1_TCLR = 3;
      ctl_global_interrupts_enable();
      ctl_task_set_priority(ctl_task_executing, sleep_priority);
    }
}

void
ctl_woken()
{
}

#endif

#endif

unsigned long
ctl_get_ticks_per_second()
{
#ifdef CTL_TASKING
  return 1000;
#else
  return 100;
#endif
}
